@@ -6,12 +6,12 @@ from threading import Semaphore, Thread, main_thread
 from typing import Any
 from uuid import uuid1
 
-from webview import (FOLDER_DIALOG, OPEN_DIALOG, SAVE_DIALOG, _settings, settings,
-                     parse_file_type, windows)
+from webview import FileDialog, _state, settings, windows
 from webview.dom import _dnd_state
 from webview.menu import Menu, MenuAction, MenuSeparator
+from webview.models import Request, Response
 from webview.screen import Screen
-from webview.util import DEFAULT_HTML, create_cookie, js_bridge_call, inject_pywebview
+from webview.util import DEFAULT_HTML, create_cookie, js_bridge_call, inject_pywebview, parse_file_type
 from webview.window import FixPoint, Window
 
 logger = logging.getLogger('pywebview')
@@ -115,6 +115,9 @@ class BrowserView:
             Gdk.Screen.get_default(), style_provider, gtk.STYLE_PROVIDER_PRIORITY_APPLICATION
         )
 
+        if window.menu:
+            logger.warning('Window specific menu is not supported on GTK')
+
         scrolled_window = gtk.ScrolledWindow()
         scrolled_window.set_policy(gtk.PolicyType.NEVER, gtk.PolicyType.NEVER)
         self.window.add(scrolled_window)
@@ -127,7 +130,7 @@ class BrowserView:
 
         self.js_bridge = BrowserView.JSBridge(window)
 
-        storage_path = _settings['storage_path'] or os.path.join(os.path.expanduser('~'), '.pywebview')
+        storage_path = _state['storage_path'] or os.path.join(os.path.expanduser('~'), '.pywebview')
 
         if not os.path.exists(storage_path):
             os.makedirs(storage_path)
@@ -135,7 +138,7 @@ class BrowserView:
         web_context = webkit.WebContext.get_default()
         self.cookie_manager = web_context.get_cookie_manager()
 
-        if not _settings['private_mode']:
+        if not _state['private_mode']:
             self.cookie_manager.set_persistent_storage(
                 os.path.join(storage_path, 'cookies'), webkit.CookiePersistentStorage.SQLITE
             )
@@ -144,17 +147,22 @@ class BrowserView:
         self.manager.register_script_message_handler('jsBridge')
         self.manager.connect('script-message-received', self.on_js_bridge_call)
 
+        self.request_headers_mutated = False
         self.webview = webkit.WebView().new_with_user_content_manager(self.manager)
         self.webview.connect('notify::visible', self.on_webview_ready)
         self.webview.connect('load_changed', self.on_load_finish)
         self.webview.connect('decide-policy', self.on_navigation)
         self.webview.connect('drag-data-received', self.on_drag_data)
+        self.webview.connect('resource-load-started', self.on_request)
+
+        if settings['IGNORE_SSL_ERRORS']:
+            web_context.set_tls_errors_policy(webkit.TLSErrorsPolicy.IGNORE)
 
         if settings['ALLOW_DOWNLOADS']:
             web_context.connect('download-started', self.on_download_started)
 
         webkit_settings = self.webview.get_settings().props
-        user_agent = settings.get('user_agent') or _settings['user_agent']
+        user_agent = settings.get('user_agent') or _state['user_agent']
         if user_agent:
             webkit_settings.user_agent = user_agent
 
@@ -184,7 +192,7 @@ class BrowserView:
             wvbg.alpha = 0.0
             self.webview.set_background_color(wvbg)
 
-        if _settings['debug']:
+        if _state['debug']:
             webkit_settings.enable_developer_extras = True
 
             if settings['OPEN_DEVTOOLS_IN_DEBUG']:
@@ -192,16 +200,16 @@ class BrowserView:
         else:
             self.webview.connect('context-menu', lambda a, b, c, d: True)  # Disable context menu
 
-        if _settings['private_mode']:
+        if _state['private_mode']:
             webkit_settings.enable_html5_database = False
             webkit_settings.enable_html5_local_storage = False
 
         self.webview.set_opacity(0.0)
         scrolled_window.add(self.webview)
 
-        if _settings['icon']:
-            self.window.set_icon_from_file(_settings['icon'])
-            self.window.set_default_icon_from_file(_settings['icon'])
+        if _state['icon']:
+            self.window.set_icon_from_file(_state['icon'])
+            self.window.set_default_icon_from_file(_state['icon'])
 
         self.pywebview_window.events.before_show.set()
 
@@ -307,20 +315,69 @@ class BrowserView:
         if 'shown' in dir(self):
             self.shown.set()
 
+    def on_response(self, resource, _):
+        response = resource.get_response()
+        headers = response.get_http_headers()
+        original_headers = self._headers_to_dict(headers)
+        url = resource.get_uri()
+
+        response_ = Response(url, response.get_status_code(), original_headers)
+        self.pywebview_window.events.response_received.set(response_)
+
+    def on_request(self, webview, resource, request):
+        if len(self.pywebview_window.events.request_sent) == 0:
+            return
+
+        if len(self.pywebview_window.events.response_received) > 0:
+            resource.connect('notify::response', self.on_response)
+
+        headers = request.get_http_headers()
+        original_headers = self._headers_to_dict(headers)
+        url = request.get_uri()
+        method = request.get_http_method()
+
+        request_ = Request(url, method, original_headers)
+        self.pywebview_window.events.request_sent.set(request_)
+
+        if (
+            request_.headers == original_headers or
+            not headers or
+            self.request_headers_mutated or
+            url != self.pywebview_window.real_url
+        ):
+            return
+
+        missing_headers = {k: v for k, v in request_.headers.items() if k not in original_headers or original_headers[k] != v}
+        extra_headers = {k: str(v) for k, v in original_headers.items() if k not in request_.headers}
+
+        for k, v in missing_headers.items():
+            headers.append(k, v)
+
+        for k in extra_headers:
+            headers.remove(k)
+
+        #headers.append('X-Handled', 'true')
+        webview.stop_loading()
+        self.request_headers_mutated = True
+        webview.load_request(request)
+
     def on_load_finish(self, webview, status):
         # Show the webview if it's not already visible
         if not webview.props.opacity:
             glib.idle_add(webview.set_opacity, 1.0)
 
-        if status == webkit.LoadEvent.FINISHED:
-            self._set_js_api()
+        if status == webkit.LoadEvent.FINISHED and not self.request_headers_mutated:
+            inject_pywebview(renderer, self.js_bridge.window)
+
+        if self.request_headers_mutated:
+            self.request_headers_mutated = False
 
     def on_download_started(self, session, download):
         download.connect('decide-destination', self.on_download_decide_destination)
 
     def on_download_decide_destination(self, download, suggested_filename):
         destination = self.create_file_dialog(
-            SAVE_DIALOG,
+            FileDialog.SAVE,
             glib.get_user_special_dir(glib.UserDirectory.DIRECTORY_DOWNLOAD),
             False,
             suggested_filename,
@@ -443,11 +500,11 @@ class BrowserView:
         return False
 
     def create_file_dialog(self, dialog_type, directory, allow_multiple, save_filename, file_types):
-        if dialog_type == FOLDER_DIALOG:
+        if dialog_type == FileDialog.FOLDER:
             gtk_dialog_type = gtk.FileChooserAction.SELECT_FOLDER
             title = self.localization['linux.openFolder']
             button = gtk.STOCK_OPEN
-        elif dialog_type == OPEN_DIALOG:
+        elif dialog_type == FileDialog.OPEN:
             gtk_dialog_type = gtk.FileChooserAction.OPEN
             if allow_multiple:
                 title = self.localization['linux.openFiles']
@@ -455,7 +512,7 @@ class BrowserView:
                 title = self.localization['linux.openFile']
 
             button = gtk.STOCK_OPEN
-        elif dialog_type == SAVE_DIALOG:
+        elif dialog_type == FileDialog.SAVE:
             gtk_dialog_type = gtk.FileChooserAction.SAVE
             title = self.localization['global.saveFile']
             button = gtk.STOCK_SAVE
@@ -471,13 +528,13 @@ class BrowserView:
         dialog.set_current_folder(directory)
         self._add_file_filters(dialog, file_types)
 
-        if dialog_type == SAVE_DIALOG:
+        if dialog_type == FileDialog.SAVE:
             dialog.set_current_name(save_filename)
 
         response = dialog.run()
 
         if response == gtk.ResponseType.OK:
-            if dialog_type == SAVE_DIALOG:
+            if dialog_type == FileDialog.SAVE:
                 file_name = (dialog.get_filename(),)
             else:
                 file_name = dialog.get_filenames()
@@ -518,8 +575,6 @@ class BrowserView:
 
             semaphore.release()
 
-        self.loaded.wait()
-
         cookies = []
         semaphore = Semaphore(0)
         glib.idle_add(_get_cookies)
@@ -528,56 +583,55 @@ class BrowserView:
         return cookies
 
     def get_current_url(self):
-        self.loaded.wait()
         uri = self.webview.get_uri()
         return uri if uri != 'about:blank' else None
 
     def load_url(self, url):
-        self.loaded.clear()
         self.webview.load_uri(url)
 
     def load_html(self, content, base_uri):
-        self.loaded.clear()
         self.webview.load_html(content, base_uri)
 
-    def evaluate_js(self, script):
+    def evaluate_js(self, script, parse_json):
         def _evaluate_js():
-            self.webview.evaluate_javascript(
-                    script=script,
-                    length=len(script),
-                    world_name=None,
-                    source_uri=None,
-                    cancellable=None,
-                    callback=_callback)
+            try:
+                self.webview.evaluate_javascript(
+                        script=script,
+                        length=len(script),
+                        world_name=None,
+                        source_uri=None,
+                        cancellable=None,
+                        callback=_callback)
+            except Exception:
+                logger.exception('Error evaluating JavaScript')
+                result_semaphore.release()
+
 
         def _callback(webview, task):
-            value = webview.evaluate_javascript_finish(task)
-            result = value.to_string() if value else None
+            nonlocal result
+            try:
+                s = script
+                value = webview.evaluate_javascript_finish(task)
+                res = self._convert_js_value(value)
 
-            if unique_id in self.js_results:
-                self.js_results[unique_id]['result'] = result
+                if parse_json and res:
+                    try:
+                        result = json.loads(res)
+                    except Exception:
+                        pass
+                else:
+                    result = res
+
+            except Exception as e:
+                logger.exception(e)
 
             result_semaphore.release()
 
-        unique_id = uuid1().hex
         result_semaphore = Semaphore(0)
-        self.js_results[unique_id] = {'semaphore': result_semaphore, 'result': None}
-
-        self.loaded.wait()
+        result = None
         glib.idle_add(_evaluate_js)
         result_semaphore.acquire()
 
-        result = self.js_results[unique_id]['result']
-        result = (
-            None
-            if result == 'undefined' or result == 'null' or result is None
-            else result
-            if result == ''
-            else json.loads(result)
-        )
-
-        del self.js_results[unique_id]
-
         return result
 
     def message_box(self, message):
@@ -591,27 +645,44 @@ class BrowserView:
         dialog.run()
         dialog.destroy()
 
-    def _set_js_api(self):
-        def create_bridge():
-            script = inject_pywebview(self.js_bridge.window, renderer)
-            self.webview.evaluate_javascript(
-                    script=script,
-                    length=len(script),
-                    world_name=None,
-                    source_uri=None,
-                    cancellable=None,
-                    callback=None)
+    def _convert_js_value(self, js_value):
+        if not js_value or js_value.is_null() or js_value.is_undefined():
+            return None
+        elif js_value.is_boolean():
+            return js_value.to_boolean()
+        elif js_value.is_number():
+            return js_value.to_double()
+        elif js_value.is_string():
+            return js_value.to_string()
+        elif js_value.is_object():
+            return json.loads(js_value.to_json(2))
+        else:
+            logger.error(f'Unsupported JavaScriptCore.Value type: {js_value}')
+            return js_value.to_string()
 
-            self.loaded.set()
+    def _headers_to_dict(self, headers):
+        def _assign(k, v):
+            headers_dict[k] = v
+        headers_dict = {}
 
-        glib.idle_add(create_bridge)
+        if headers:
+            headers.foreach(_assign)
+        return headers_dict
 
 
 def setup_app():
-    # MUST be called before create_window and set_app_menu
+    def set_menubar(app):
+        app.set_menubar(app_menu)
+
     global _app
-    if _app is None:
-        _app = gtk.Application.new(None, 0)
+    if _app is not None:
+        return
+
+    _app = gtk.Application.new(None, 0)
+
+    if _state['menu']:
+        app_menu = create_menu(_state['menu'])
+        _app.connect('startup', set_menubar)
 
 
 def create_window(window):
@@ -790,15 +861,7 @@ def create_confirmation_dialog(title, message, uid):
     return result
 
 
-def set_app_menu(app_menu_list):
-    """
-    Create a custom menu for the app bar menu (on supported platforms).
-    Otherwise, this menu is used across individual windows.
-
-    Args:
-        app_menu_list ([webview.menu.Menu])
-    """
-    global _app_actions
+def create_menu(app_menu_list):
 
     def action_callback(action, parameter):
         function = _app_actions.get(action.get_name())
@@ -838,17 +901,13 @@ def set_app_menu(app_menu_list):
 
         supermenu.append_submenu(title, m)
 
-    global _app
-
     menubar = Gio.Menu()
 
     for app_menu in app_menu_list:
         create_submenu(app_menu.title, app_menu.items, menubar)
 
-    def set_menubar(app):
-        app.set_menubar(menubar)
+    return menubar
 
-    _app.connect('startup', set_menubar)
 
 
 def get_active_window():
@@ -889,11 +948,11 @@ def create_file_dialog(dialog_type, directory, allow_multiple, save_filename, fi
     return file_names[0]
 
 
-def evaluate_js(script, uid):
+def evaluate_js(script, uid, parse_json=True):
     i = BrowserView.instances.get(uid)
 
     if i:
-        return i.evaluate_js(script)
+        return i.evaluate_js(script, parse_json)
 
 
 def get_position(uid):
@@ -937,7 +996,7 @@ def get_screens():
     n = display.get_n_monitors()
     monitors = [Gdk.Display.get_monitor(display, i) for i in range(n)]
     geometries = [Gdk.Monitor.get_geometry(m) for m in monitors]
-    screens = [Screen(geom.width, geom.height, geom) for geom in geometries]
+    screens = [Screen(geom.x, geom.y, geom.width, geom.height, geom) for geom in geometries]
 
     return screens
 
